/**
 SequenceMiner: multiple anomaly classifier
 
 @author: Eugene Turkov, {eturkov@email.arc.nasa.gov}
 @version: 1.0
 
 NASA Ames Research Center
 Computational Sciences Division
 Intelligent Data Understanding Group (IDU)
 Moffett Field, CA 94035
 */

package turkov.math.algorithms;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
import java.lang.Math;
import turkov.util.TUSet;
import turkov.lang.TUStringBuilder;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Arrays;
import turkov.io.TUTextFileWriter;
import java.io.IOException;
import turkov.util.TUCollectionUtil;
import turkov.io.TUCSVFile;
import java.util.Vector;

public abstract class TUClusteringMethods
{
    /*
        Compared to the well-known kmeans algorithm, PAM has the following features:
     
        It operates on the dissimilarity matrix of the given data set or when it is
        presented with an n � p data matrix, the algorithm first computes a dissimilarity matrix.
     
        It is more robust, because it minimizes a sum of dissimilarities instead
        of a sum of squared Euclidean distances.
     
        It provides a novel graphical display, the silhouette plot, which allows
        the user to select the optimal number of clusters.
     
        @param numberOfClustersDesired the number of clusters that this method should
        produce.
     
        @param data a complex array of data items to cluster. An ideal example would be
        an array of Character arrays. Such a structure is neccessary due to the complexity
        of some of the assisting algorithms which this method uses.
     
        @param ditanceMethodToUse this algorithm needs to be able to calculate
        distances between each individual data item. Available distance measures can be
        found at constants in the DistanceMesuares class
     
        @param numberOfIterations the number of times this algorithm should be run
        in an attempt to determine ideal clusters.
     
        @return an array of clusters.
     */
    
    public static TUCluster[] getPAMClusters(int numberOfClustersDesired,
            TUClusterableItem[] data, String distanceMethodToUse
            ,int numberOfIterations, final TUSet medoids)
    {
    	
   	
        if(numberOfClustersDesired < 1)
            throw new UnsupportedOperationException("Can't cluster data into zero clusters.");
        
        Random randomNumberGenerator = new Random();
        TUCluster[] clusters = new TUCluster[numberOfClustersDesired];
        for(int i = 0; i < numberOfClustersDesired; i++)
            clusters[i] = new TUCluster();
        
        TUClusterableItem[] suggestedMedoids = new TUClusterableItem[medoids.size()];
               
            /* step 2, do partitioning -> go through each data value and assign
                it to the cluter which is closest (based on medoid) */
        
        for(int i = 0; i < medoids.size(); i++)
            suggestedMedoids[i] = (TUClusterableItem)medoids.get(i);
 
        HashMap<Object, ArrayList<Integer>> seqIndex = null;
        int len2 = 0;
        Object[] seq1 = null;
        Object[] seq2 = null;
        for(int iteration = 0; iteration < numberOfIterations; iteration ++)
        {
            clusters = new TUCluster[numberOfClustersDesired];
            for(int k = 0; k < clusters.length; k++)
                clusters[k] = new TUCluster();
            
            for(int i = 0; i < data.length; i++)
            {
                double shortestDistance = Double.MAX_VALUE;
                int medoidIdWithShortestDistance = 0;
                if(distanceMethodToUse.equals(TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE))
                {
                	seq2 = (Object[])data[i].getValue();
                	len2 = seq2.length;
                	seqIndex = TULCS.getSequenceIndex(seq2);
                }
                
                // go through each medoid and compare..
                for(int j = 0; j < medoids.size(); j++)
                {
                	double distance=0;
                	if(distanceMethodToUse.equals(TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE))
                	{
                		if((Object[])((TUClusterableItem)medoids.get(j)).getValue()!=null)
                			seq1 = ((Object[])((TUClusterableItem)medoids.get(j)).getValue());
                		else
                			System.out.println("Null Object. ID = "+j);
                		
                		distance = TUDistanceMeasures.getDistance(seq1, seqIndex, len2, distanceMethodToUse);
                	}
                	else
                	{
                    distance = TUDistanceMeasures.getDistance
                            (    ((Object[])((TUClusterableItem)medoids.get(j)).getValue()) ,  ((Object[])data[i].getValue()),
                            distanceMethodToUse);
                	}

                    if(distance <= shortestDistance)
                    {
                        shortestDistance = distance;
                        medoidIdWithShortestDistance = j;
                    }
                }
                
                // before continuing, add this data element to the appropriate cluster
                clusters[medoidIdWithShortestDistance].add(data[i]);
            }
            
            // now the above loops is done and we should have some clusters.
            
            // step 3, for each cluster get the suggested medoids
            for(int i = 0; i < clusters.length; i++)
            {
                if(suggestedMedoids[i] == null)
                    continue;
                /*
                if(suggestedMedoids[i].equals(clusters[i].getMedoid(distanceMethodToUse)))
                    break;
                 */
                
                //else
                suggestedMedoids[i] = (TUClusterableItem)clusters[i].getMedoid(distanceMethodToUse);
            }
        }
        
        
        // run the below code to set outlier scores for all items that have been clustered
        for(int m = 0; m < clusters.length; m++)
        {
            
            TUCluster cluster = clusters[m];
            
            if(cluster.size() == 0)
                continue;
            
            Object[] a = (Object[])(((TUClusterableItem)clusters[m]
                        .getMedoid(distanceMethodToUse)).getValue());
                
            for(int j = 0; j < cluster.size(); j++)
            {
                
                Object[] b = (Object[])(((TUClusterableItem)cluster.get(j)).getValue());
                
                double distance = TUDistanceMeasures
                        .getDistance(a,b, distanceMethodToUse);
                
                ((TUClusterableItem)cluster.get(j)).setOutlierScore(distance);
            }
        }
        
        /*
        System.err.println(System.getProperty("line.separator") + "PAM Clustering Quality: "
                + TUClusteringMethods.getClusteringQuality(clusters,
                TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE));
         */
        /*
        for(int i = 0; i < clusters.length; i++)
        {
            TUCluster cluster = clusters[i];
            /*
            String output = "The size of PAM cluster " + (i + 1) + " is: "
                        + cluster.size();
         */
            /*
            if(i == clusters.length - 1)
                output += System.getProperty("line.separator");
             */
        
            /*
            System.err.println(output);
             */
        /*
        }
         */
        return clusters;
    }
    
    private static TUClusterableItem[] getClusterableItems(String[] sequences)
    {
        /* using a vector initially because we don't know how many sequences we
        will be working with */
        TUClusterableItem[] clusterableItems = null;
        Vector temp = new Vector();
 
        for(int i = 0; i < sequences.length; i++)
        {
            sequences[i] = sequences[i].trim();
            String[] sequenceData = sequences[i].split(" ");
            
            /*
            System.out.println("just read in: "
            + TUStringBuilder.concatenate(sequenceData, "<space>"));
             */
            
            if(sequenceData.length == 0
                    || (sequenceData.length == 1
                    && sequenceData[0].equals("")))
                continue;
            
            Vector tempSequence = new Vector();
            for(int j = 0; j < sequenceData.length; j++)
            {
                try
                {
                    if(sequenceData[j].trim().equals(""))
                        continue;
                    
                    //System.out.println("adding: " + sequenceData[j]);
                    tempSequence.add(String.valueOf(sequenceData[j]));
                    //sequence[j] = Integer.valueOf(sequenceData[j]);
                }
                catch(NumberFormatException e)
                {
                    System.err.println("Read in a bad element in the input file" +
                            " at line: " + i + " check the: " + (j + 1) + "th item.");
                }
                
            }
            
            /* convert the vector back over to an array */
            String[] sequence = new String[tempSequence.size()];
            for(int k = 0; k < tempSequence.size(); k++)
                sequence[k] = (String)tempSequence.get(k);
            
            temp.add(new TUClusterableItem(sequence, i));
        }
        
        /* convert the vector back over to an array */
        clusterableItems = new TUClusterableItem[temp.size()];
        for(int i = 0; i < temp.size(); i++)
            clusterableItems[i] = (TUClusterableItem)temp.get(i);
        
        return clusterableItems;
    }    
    
    public static TUCluster[] getClaraClusters(final TUCSVFile predeterminedClustersCSVFile,
            final File inputFile)
            throws IOException
    {
        final int SEQUENCE_ID = 0;
        final int OWNER_CLUSTER_ID = 1;
        final int OUTLIER_SCORE = 2;
        final int RANK = 3;
        final int IS_OUTLIER = 4;
        final int TOTAL_NUMBER_OF_COLUMNS = 5;

        
        String[] sequences = TUStringBuilder.generateCollectionFromFile(inputFile);
        
        
        TUClusterableItem[] clusterableItems = getClusterableItems(sequences);  
        
        final int rows = predeterminedClustersCSVFile.getRows();            

        Object[][] clusteringInfo = new Object[rows][TOTAL_NUMBER_OF_COLUMNS];
        int highestClusterId = 0;
        for(int row = 0; row < rows; row++)
        {
            for(int column = 0; column < TOTAL_NUMBER_OF_COLUMNS; column++)
            {
                if(column == SEQUENCE_ID || column == OWNER_CLUSTER_ID
                        || column == RANK || column == IS_OUTLIER)
                    clusteringInfo[row][column]
                        = predeterminedClustersCSVFile.getInteger(row, column);
                else if(column == OUTLIER_SCORE)
                    clusteringInfo[row][column]
                        = predeterminedClustersCSVFile.getDouble(row, column);
                
                else
                {
                    // do nothing
                }
                
                if(column == OWNER_CLUSTER_ID)
                {
                    int ownerClusterId = ((Integer)clusteringInfo[row][column])
                        .intValue();
                    if(ownerClusterId > highestClusterId)
                        highestClusterId = ownerClusterId;
                }
                
                else if(column == RANK)
                {                                   
                    clusterableItems[((Integer)clusteringInfo[row][SEQUENCE_ID])
                    .intValue() - 1].setRank(((Integer)clusteringInfo[row][RANK])
                    .intValue());
                    
                    clusterableItems[((Integer)clusteringInfo[row][SEQUENCE_ID])
                    .intValue() - 1].setOutlierScore(((Double)clusteringInfo[row]
                            [OUTLIER_SCORE]).doubleValue());        
                    
                    clusterableItems[((Integer)clusteringInfo[row][SEQUENCE_ID])
                    .intValue() - 1].setOwnerClusterId(((Integer)clusteringInfo[row][OWNER_CLUSTER_ID])
                    .intValue() - 1);                                        
                }
                
                else if(column == IS_OUTLIER)
                {
                    int isOutlier = ((Integer)clusteringInfo[row][IS_OUTLIER]).intValue();
                    if(isOutlier == 1)
                    {
                        clusterableItems[((Integer)clusteringInfo[row][SEQUENCE_ID])
                        .intValue() - 1].setCustomData("isOutlier", "true");
                    }
                }
            }          
        }
        
        TUCluster[] clusters = new TUCluster[highestClusterId];
        for(int i = 0; i < clusters.length; i++)
            clusters[i] = new TUCluster();
        
        for(int row = 0; row < clusteringInfo.length; row++)
        {
            clusters[((Integer)clusteringInfo[row][OWNER_CLUSTER_ID])
                .intValue() - 1].add(clusterableItems[((Integer)clusteringInfo[row]
                    [SEQUENCE_ID]).intValue() - 1]);            
        }                
        
        return clusters;
    }
    
    public static TUCluster[] getClaraClusters(int numberOfClustersDesired,
            Object[] data, String distanceMethodToUse, int numberOfPAMIterations,
            int numberOfClaraIterations, double sampleSizePercentage,int randomSeed)
    {
    	
    	double bestQualitySoFar = Double.POSITIVE_INFINITY;
        TUCluster[] bestPartialClustersSoFar = null;
        
        int bestIterationSoFar = -1;
        
        int sampleSize = (int)Math.ceil(Math.max(Math.ceil(sampleSizePercentage
                * (double)data.length), (double)numberOfClustersDesired));
        
        
        TUCluster[] partialClusters = null;
        
        //Set up Random seed if not defined.
        Random randomNumberGenerator = new Random();
        if(randomSeed==-1) 
        {
        	int undefinedSeed = java.lang.Math.abs(randomNumberGenerator.nextInt());
        	randomNumberGenerator.setSeed(undefinedSeed);
        	System.out.println("Using random seed =" + undefinedSeed);
        }	
        else
        {
        	randomNumberGenerator = new Random(randomSeed);
        	System.out.println("Using specified random seed =" + randomSeed);
        }
        

        /*
        if(randomSeed==-1) {randomNumberGenerator = new Random();}	
        else{randomNumberGenerator = new Random(randomSeed);}	
       */
        
        
        TUSet medoids = new TUSet();
        
        // step 1, pick the as many medoids as clusters needed
        while(medoids.size() < numberOfClustersDesired)
        {
            //System.out.println("data.length: " + data.length);
            int medoidId = randomNumberGenerator.nextInt(data.length);
            System.out.println("Medoid ID ="+medoidId);
            medoids.add(data[medoidId]);
        }
        
        for(int i = 0; i < numberOfClaraIterations; i++)
        {
            
            Object[] dataSubset = TUCollectionUtil.getRandomSubset(data, sampleSize,randomNumberGenerator.nextInt());
            TUClusterableItem[] actualDataSubset = new TUClusterableItem[dataSubset.length];
            for(int m = 0; m < dataSubset.length; m++)
            {
                actualDataSubset[m] = (TUClusterableItem)dataSubset[m];
            }
            
            partialClusters = getPAMClusters(numberOfClustersDesired, actualDataSubset,
                    distanceMethodToUse, numberOfPAMIterations, medoids);
            
            double clusteringQuality = getClusteringQuality(partialClusters, distanceMethodToUse);
            
            System.err.println("Iteration: " + (i + 1));
            System.err.println("Quality: " + clusteringQuality);
            
            for(int k = 0; k < partialClusters.length; k++)
            {
                System.err.println("Cluster: " + (k + 1) + ", size: " +
                        partialClusters[k].size() + ", quality: "
                        + partialClusters[k].getQuality(distanceMethodToUse));
                

            }            

            if(clusteringQuality < bestQualitySoFar)
            {
                // fix any clusters that have zero items (medoid purposes)
                for(int z = 0; z < partialClusters.length; z++)
                {
                    if(partialClusters[z].getMedoid(distanceMethodToUse) == null)
                    {
                        Object medoidCandidate = data[randomNumberGenerator.nextInt(data.length)];
                        for(int h = 0; h < partialClusters.length; h++)
                        {                            
                            if(partialClusters[h].getMedoid(distanceMethodToUse) != null
                                    && partialClusters[h]
                                    .getMedoid(distanceMethodToUse)
                                    .equals(medoidCandidate))
                            {
                                // pick a new candidate
                                medoidCandidate = data[randomNumberGenerator.nextInt(data.length)];
                                // reset counter
                                h = -1;
                                continue;
                            }
                            
                        }
                        
                        partialClusters[z].add(medoidCandidate);
                    }
                }
                
                
                
                medoids = new TUSet();
                for(int k= 0; k < partialClusters.length; k++)
                    medoids.add(partialClusters[k].getMedoid(distanceMethodToUse));
                
                bestQualitySoFar = clusteringQuality;
                bestPartialClustersSoFar = partialClusters;
                bestIterationSoFar = i;

            }                        
        }
        
        System.err.println(System.getProperty("line.separator")
        + "Clustering configuration selected: CLARA iteration "
                + (bestIterationSoFar + 1));
        
        TUCluster[] fullClusters = getFullClusters(bestPartialClustersSoFar, data, distanceMethodToUse);
        
        for(int m = 0; m < fullClusters.length; m++)
        {
            System.err.println("\tfull cluster " + (m + 1) + " size is: " + fullClusters[m].size());
        }
        
        for(int m = 0; m < fullClusters.length; m++)
        {
            
            TUCluster cluster = fullClusters[m];
            System.err.println("\tcomputing outlier scores for cluster " + (m + 1) + ", which has size " + cluster.size());
            
            Object[] a = null;
            try{
            	a = (Object[])(((TUClusterableItem)fullClusters[m].getMedoid(distanceMethodToUse)).getValue());
            	}
            catch(java.lang.NullPointerException e)
            {
            	/*When cluster size is 0 null pointer is thrown when getting Mediod since no Mediod exists. 
            	 Solution catch exception and do nothing.*/ 
            }
            
            
            for(int j = 0; j < cluster.size(); j++)
            {
                //System.err.println("\t\tcomputing outlier scores for element " + j);                
                //System.err.println("\t\t\t\tDone retrieving medoid of this cluster");
                Object[] b = (Object[])(((TUClusterableItem)cluster.get(j)).getValue());
                
                double distance = TUDistanceMeasures.getDistance
                        (a,b, distanceMethodToUse);
                //System.err.println("\t\t\tDone computing outlier score for element " + j);
                ((TUClusterableItem)cluster.get(j)).setOutlierScore(distance);
            }
        }
        
        return fullClusters;
    }
    
    
    public static TUCluster[] getFullClusters(TUCluster[] partialClusters, Object[] data,
            String distanceFunction)
    {
        TUCluster[] result = new TUCluster[partialClusters.length];
        for(int i = 0; i < result.length; i++)
            result[i] = new TUCluster();
        
        for(int i = 0; i < data.length; i++)
        {
            if(partialClusters.length == 0)
                continue;
            
            double minDistanceSoFar = Double.POSITIVE_INFINITY;
            int idOfPartialClusterWithBestMatchingMedoid = -1;
            
            Object[] b = (Object[])(((TUClusterableItem)data[i]).getValue());            
            
            for(int j = 0; j < partialClusters.length; j++)
            {
                if(partialClusters[j].getMedoid(distanceFunction) == null)
                    continue;
                
                Object[] a = (Object[])(((TUClusterableItem)partialClusters[j].getMedoid(distanceFunction)).getValue());                
                double distance = TUDistanceMeasures.getDistance
                        (a,b, distanceFunction);
                
                //System.err.println("caluclated distance between data point " + i + " and " + j + ": " + distance);
                
                if(distance < minDistanceSoFar)
                {
                    minDistanceSoFar = distance;
                    idOfPartialClusterWithBestMatchingMedoid = j;
                }
            }
            
            try
            {
                //((TUClusterableItem)data[i]).setOutlierScore(minDistanceSoFar);
                result[idOfPartialClusterWithBestMatchingMedoid].add(data[i]);
            }
            catch(ArrayIndexOutOfBoundsException e)
            {
                System.err.println("i: " + i);
                System.err.println("partial cluster id: " + idOfPartialClusterWithBestMatchingMedoid);
                System.err.println("partial clusters length: " + partialClusters.length);
                System.err.println("Min dist so far: " + minDistanceSoFar);
                e.printStackTrace();
            }
        }
        
        return result;
    }
    
    public static double getClusteringQuality(TUCluster[] clusters, String method)
    {
        if(clusters == null)
            return 0;
     
        int totalNumberOfItems = 0;
        for(int i = 0; i < clusters.length; i++)
        {
            totalNumberOfItems += clusters[i].size();
        }
        
        double result = 0;
        for(int i = 0; i < clusters.length; i++)
        {
            if(clusters[i].size() == 0)
                continue;
            
            result += clusters[i].getQuality(method) * (clusters[i].size()) / totalNumberOfItems;
        }        
        
        return result;
    }
}
