/**
 * @author Eugene Turkov
 * @author <a href="mailto:eugene.turkov@nasa.gov">eugene.turkov@nasa.gov</a>
 * @version: 1.0
 * @since SDK1.5
 * 
 * NASA Ames Research Center Computational Sciences Division Intelligent Data
 * Understanding Group (IDU) Moffett Field, CA 94035
 */
package turkov.util;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import turkov.io.TUTextFileWriter;
import turkov.lang.TUObjectExtension;
import turkov.math.TUMaxMin;
import turkov.math.algorithms.TUCluster;
import turkov.math.algorithms.TUClusterableItem;
import turkov.math.algorithms.TUDistanceMeasures;
import turkov.math.algorithms.TULCS;

public abstract class TUClusterUtil
{
	/**
	 This method iterates through a collection of clusters and sets the rank
	 values of each element in each cluster.
	 
	 @param clusters the collection of clusters upon which to operate.
	 */
	public static void setRankValues(TUCluster[] clusters)
	{
		if (clusters == null)
			return;

		int count = 0;

		for (int i = 0; i < clusters.length; i++)
		{
			count += clusters[i].size();
		}

		TUClusterableItem[] result = new TUClusterableItem[count];

		count = 0;

		for (int i = 0; i < clusters.length; i++)
		{
			TUCluster cluster = clusters[i];
			for (int j = 0; j < cluster.size(); j++)
			{
				TUClusterableItem item = (TUClusterableItem) cluster.get(j);
				item.setCompareMethod(item.COMPARE_BY_OUTLIER_SCORE);
				result[count] = item;
				count++;
			}
		}

		Arrays.sort(result);

		for (int i = 0; i < result.length; i++)
		{
			result[i].setRank(result.length - 1 - i);
			//System.out.println("got a rank of: " + result[i].getRank() + " and an outlier score of: " + result[i].getOutlierScore());            
			result[i].setCompareMethod(TUClusterableItem.COMPARE_BY_ID);
		}
	}

	/**
	 This is a customized output method for clusters.
	 
	 @param clusters the clusters to be output.
	 @param output file the file to which clusters should be output to.
	 @param sort whether or not the clusters should be sorted.
	 @param outlierFraction the fraction of outliers.
	 */
	public static void writeOutputCSVFile(TUCluster[] clusters,
			File outputFile, boolean sort, double outlierFraction,
			boolean onlyPrintNewSequences, Vector <String> sequencesDefId) throws IOException
	{
		if (clusters == null)
			return;

		File outputCSVFile = outputFile;
		int totalNumberOfItems = 0;

		for (int i = 0; i < clusters.length; i++)
			totalNumberOfItems += clusters[i].size();

		TUClusterableItem[] allItems = new TUClusterableItem[totalNumberOfItems];
		TUClusterableItem[] duplicateAllItemsForScoreSorting = null;
		int cutoffRankInclusive = (int) Math.ceil(totalNumberOfItems
				* outlierFraction) - 1;

		for (int i = 0; i < clusters.length; i++)
		{
			for (int j = 0; j < clusters[i].size(); j++)
			{
				allItems[totalNumberOfItems - 1] = (TUClusterableItem) clusters[i]
						.get(j);
				allItems[totalNumberOfItems - 1].setOwnerClusterId(i);
				totalNumberOfItems--;
			}
		}

		StringBuffer stringBuffer = new StringBuffer();

		if (onlyPrintNewSequences)
		// create a separate list of items and sort it based on outlier score
		// to give new rank
		{
			Vector newItems = new Vector();
			for (int i = 0; i < allItems.length; i++)
			{
				if (allItems[i].getCustomData(SUPPLEMENTAL_ITEM) != null)
				{
					newItems.add(allItems[i]);
				}
			}

			allItems = new TUClusterableItem[newItems.size()];
			duplicateAllItemsForScoreSorting = new TUClusterableItem[allItems.length];

			for (int i = 0; i < newItems.size(); i++)
			{
				allItems[i] = (TUClusterableItem) newItems.get(i);
				duplicateAllItemsForScoreSorting[i] = allItems[i];
			}
		}

		if (sort)
			Arrays.sort(allItems);

		if (onlyPrintNewSequences)
		{
			for (int i = 0; i < duplicateAllItemsForScoreSorting.length; i++)
			{
				duplicateAllItemsForScoreSorting[i]
						.setCompareMethod(TUClusterableItem.COMPARE_BY_OUTLIER_SCORE);
			}

			Arrays.sort(duplicateAllItemsForScoreSorting);

			for (int i = 0; i < duplicateAllItemsForScoreSorting.length; i++)
			{
				duplicateAllItemsForScoreSorting[i]
						.setRank((duplicateAllItemsForScoreSorting.length - 1)
								- i);
			}
		}

		for (int i = 0; i < allItems.length; i++)
		{
			if (onlyPrintNewSequences
					&& allItems[i].getCustomData(SUPPLEMENTAL_ITEM) == null)
				continue;

	        //Use defined sequence ID (if it exists) instead of row number
			if(sequencesDefId.size()>1)
				stringBuffer.append(sequencesDefId.get(allItems[i].getId()));
			
			else
				stringBuffer.append(allItems[i].getId() + 1);
			
			stringBuffer.append(",");
			stringBuffer.append(allItems[i].getOwnerClusterId() + 1);
			stringBuffer.append(",");
			stringBuffer.append(allItems[i].getOutlierScore());
			stringBuffer.append(",");
			stringBuffer.append(allItems[i].getRank() + 1);
			stringBuffer.append(",");

			if (allItems[i].getRank() <= cutoffRankInclusive)
				stringBuffer.append("1");

			else
				stringBuffer.append("0");

			//if(i + 1 < allItems.length)
			stringBuffer.append(System.getProperty("line.separator"));
		}

		TUTextFileWriter.createFile(outputCSVFile, new String[] { stringBuffer
				.toString() });
	}

	/**
	 Non-Essential Elements algorithm from the paper entitled "Anomaly
	 Detection and Diagnosis Algorithms for Discrete Symbol Sequences with
	 Applications to Airline Safety" by Suratna Budalakoti.
	 
	 This algorithm detects extra elements from a sequence given a cluster of
	 sequences.     
	 
	 @param clusterableItem and item which can be part of a cluster.
	 @param cluster a cluster containing clusterable items (sequences of elements).
	 @param uniqueElementsToIds a map where the key set is the set of unique elements
	 and the value set contains the corresponding element ids.
	 @param idsToUniqueElements a map where the kye set is the set of unique ids
	 and the value set contains the corresponding elements.
	 @return an array of points. For each point, x specifies a location in the
	 outlier sequence (clusterableItem) and y specifies an id of an element that
	 should be inserted before position x in the outlier sequence (clusterableItem).
	 */
	public static TUSequenceAlteration[] getNonEssentialElements(
			final Object[] outlierSequence, final TUCluster cluster,
			final HashMap uniqueElementsToIds,
			final HashMap idsToUniqueElements, int sequenceId)
	{
		//System.err.println("getNonEssentialElements called on sequence with id: " + sequenceId);
		Iterator iterator = uniqueElementsToIds.keySet().iterator();

		/*
		while(iterator.hasNext())
		{
		    Object key = iterator.next();
		    System.err.println(key + " --> " + uniqueElementsToIds.get(key));
		    
		    if(!iterator.hasNext())
		        System.err.println("");
		}
		 */

		final int totalNumberOfUniqueSymbols = uniqueElementsToIds.size();

		Vector missingCharacterIndeciesAndIds = new Vector();

		double[] b = new double[outlierSequence.length];

		final Object[] consensusSequence = (Object[]) ((TUClusterableItem) cluster
				.getMedoid(TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE))
				.getValue();

		Vector result = new Vector();

		double score = 0;
		double oldScore = 0;
		//System.err.println("\n\n\nJUST STARTED FINDING EXTRA ELEMENTS FOR THIS SEQUENCE\n") ;       
		for (int i = 0; i < cluster.size(); i++)
		{

			if (((TUClusterableItem) cluster.get(i))
					.getCustomData(turkov.util.TUClusterUtil.SUPPLEMENTAL_ITEM) != null)
				continue;
			//System.out.println("processing sequence " + i + "...");
			// here, k are the global id of elements
			final Object[] currentSequence = (Object[]) ((TUClusterableItem) cluster
					.get(i)).getValue();

			final Object[] longestCommonSubsequence = TULCS
					.getLongestCommonSubsequence(outlierSequence,
							currentSequence);

			// to determine mappings below
			int currentPosition = 0;

			int[] lcsToCurrentSequenceIndices = new int[longestCommonSubsequence.length];

			int[] lcsToOutlierSequenceIndices = new int[longestCommonSubsequence.length];

			int[] outlierSequenceToLcsIndices = new int[outlierSequence.length];

			// first determine lcsToCurrentSequenceMapping
			for (int j = 0; j < longestCommonSubsequence.length; j++)
			{
				lcsToCurrentSequenceIndices[j] = TUCollectionUtil.getPosition(
						longestCommonSubsequence[j], currentSequence,
						currentPosition);

				currentPosition = lcsToCurrentSequenceIndices[j] + 1;
			}

			//System.out.println("this is the lcs");
			//TUStringBuilder.printCollection(longestCommonSubsequence);

			//reset currentPosition
			currentPosition = 0;

			// second, determine outlierSequenceToLcsMapping
			for (int j = 0; j < outlierSequence.length; j++)
			{
				outlierSequenceToLcsIndices[j] = -1;
			}

			for (int lcsIndex = 0; lcsIndex < longestCommonSubsequence.length; lcsIndex++)
			{

				outlierSequenceToLcsIndices[TUCollectionUtil.getPosition(
						longestCommonSubsequence[lcsIndex], outlierSequence,
						currentPosition)] = lcsIndex;
				/*
				System.out.println("outlier item: " + TUCollectionUtil.getPosition(longestCommonSubsequence[lcsIndex],
				        outlierSequence, currentPosition) + " --> lcs item: " + lcsIndex);
				 */
				currentPosition = outlierSequenceToLcsIndices[currentPosition] + 1;
			}

			currentPosition = 0;

			// third, determine lcsToOutlierSequenceMapping                       
			for (int j = 0; j < longestCommonSubsequence.length; j++)
			{
				lcsToOutlierSequenceIndices[j] = TUCollectionUtil.getPosition(
						longestCommonSubsequence[j], outlierSequence,
						currentPosition);

				currentPosition = lcsToOutlierSequenceIndices[j] + 1;
			}

			for (int j = 0; j < outlierSequence.length; j++)
			{
				// add an element to the non-essential items list because
				// it doesn't belong in the lcs
				if (outlierSequenceToLcsIndices[j] >= 0)
				{
					b[j] += ((double) longestCommonSubsequence.length + 1)
							/ ((double) currentSequence.length + 1);
				}

				else
				{/*
				                    System.out.println("j: " + j);
				                   System.out.println("found bad item: " + outlierSequence[j]);
				 */
				}

			}

			int lengthOfLongestCommonSubsequence = -1;

			final TUClusterableItem currentSequenceObject = ((TUClusterableItem) cluster
					.get(i));

			if (currentSequenceObject.getCustomData(DISTANCE_FROM_MEDOID) == null)
			{
				lengthOfLongestCommonSubsequence = TULCS
						.getLengthOfLongestCommonSubsequence(consensusSequence,
								currentSequence);

				currentSequenceObject.setCustomData(DISTANCE_FROM_MEDOID,
						new Integer(lengthOfLongestCommonSubsequence));
			}

			else
				lengthOfLongestCommonSubsequence = ((Integer) currentSequenceObject
						.getCustomData(DISTANCE_FROM_MEDOID)).intValue();

			score += (longestCommonSubsequence.length * lengthOfLongestCommonSubsequence)
					/ (Math.sqrt(outlierSequence.length) * currentSequence.length);

		}

		/*
		for(int i = 0; i < b.length; i++)
		    System.out.println("b[" + i + "] = " + b[i]);
		 */

		//System.out.println("initial score: " + score);
		oldScore = score;

		double tempMax = TUMaxMin.getMaxValue(b);

		//System.err.println("score: " + score);

		/*
		for(int i = 0; i < b.length; i++)
		{
		    System.err.println("b[" + i + "] = " + b[i] + " and outlierSequence[" + i + "] = " + outlierSequence[i]);
		}
		System.err.println("---\n\n");         

		System.err.println("Now removing characters."); 
		 */
		while (true)
		{
			/* step 3 -- determine tempMin, tempMax, numberOfTempMinInstances,
			   and the instances themselves */

			double tempMin = TUMaxMin.getMinValue(b);
			//System.out.println("temp min is: " + tempMin);
			int numberOfTempMinInstances = TUCollectionUtil
					.getNumberOfOccurences(tempMin, b);
			if (tempMin != 0 && numberOfTempMinInstances == b.length)
			{
				//System.out.println("here");
				break;
			}

			//System.out.println("number of tempmin instances is : " + numberOfTempMinInstances);
			int[] tempMinIndecies = new int[numberOfTempMinInstances];

			int count = 0;
			for (int i = 0; i < b.length; i++)
			{
				if (b[i] == tempMin)
				{
					tempMinIndecies[count] = i;
					count++;
				}
			}

			/*
			System.err.println("k: " + tempMinIndecies.length);
			System.err.println("f (old): " + score);
			System.err.println("b min: " + tempMin);
			
			Object[] stuff = new Object[tempMinIndecies.length];
			
			System.err.print("Now Removing: ");
			 
			for(int i = 0; i < tempMinIndecies.length; i++)
			{
			    stuff[i] = outlierSequence[tempMinIndecies[i]];
			   // System.err.print(stuff[i] + " ");
			}
			                        System.err.println("");
			 */

			score = (1 / Math.sqrt(outlierSequence.length
					- tempMinIndecies.length))
					* (Math.sqrt(outlierSequence.length) * oldScore - tempMinIndecies.length
							* tempMin);

			//System.err.println("f (new): " + score);            

			if (score <= oldScore)
			{
				//System.out.println("hey stupid!");
				break;
			}

			/*
			score = (1 / Math.sqrt(outlierSequence.length - tempMinIndecies.length))
			 * ((1 / Math.sqrt(outlierSequence.length))
			 * oldScore
			- tempMinIndecies.length
			 * tempMin);            
			 */
			if (new Double(score).isNaN())
				score = Double.POSITIVE_INFINITY;

			/*
			System.out.println("oldScore: " + oldScore);
			System.out.println("new score: " + score);
			 */

			for (int i = 0; i < tempMinIndecies.length; i++)
			{
				{
					result.add(new TUSequenceAlteration(tempMinIndecies[i],
							outlierSequence[tempMinIndecies[i]],
							b[tempMinIndecies[i]]));

					//System.out.println("adding to result " + outlierSequence[tempMinIndecies[i]]);

				}

				b[tempMinIndecies[i]] = tempMax + 1;
			}

			oldScore = score;
		}

		TUSequenceAlteration[] finalResult = new TUSequenceAlteration[result
				.size()];

		for (int i = 0; i < result.size(); i++)
			finalResult[i] = (TUSequenceAlteration) result.get(i);

		//System.err.println("the final number of extra elements is: " + finalResult.length);

		return finalResult;
	}

	/**
	 *     Algorithm 5 from the paper entitled "Anomaly Detection and Diagnosis
	 *     Algorithms for Discrete Symbol Sequences with Applications to Airline
	 *     Safety" by Suratna Budalakoti
	 *     
	 *     This algorithm detects missing elements from a sequence given a cluster of
	 *     sequences.
	 *     
	 * 
	 * @param clusterableItem and item which can be part of a cluster.
	 * @param cluster a cluster containing clusterable items (sequences of elements).
	 * @param uniqueElementsToIds a map where the key set is the set of unique elements
	 *     and the value set contains the corresponding element ids.
	 * @param idsToUniqueElements a map where the kye set is the set of unique ids
	        and the value set contains the corresponding elements.
	 * @return an array of points. For each point, x specifies a location in the
	 *     outlier sequence (clusterableItem) and y specifies an id of an element that
	 *     should be inserted before position x in the outlier sequence (clusterableItem).
	 */
	public static TUSequenceAlteration[] getMissingElements(
			final Object[] outlierSequence, final TUCluster cluster,
			final HashMap uniqueElementsToIds, final HashMap idsToUniqueElements)
	{
		final int totalNumberOfUniqueSymbols = uniqueElementsToIds.size();

		Vector missingCharacterIndeciesAndIds = new Vector();

		double[][] b = new double[outlierSequence.length + 1][totalNumberOfUniqueSymbols];
		final Object[] consensusSequence = (Object[]) ((TUClusterableItem) cluster
				.getMedoid(TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE))
				.getValue();

		double score = 0;
		double oldScore = 0;

		for (int i = 0; i < cluster.size(); i++)
		{
			if (((TUClusterableItem) cluster.get(i))
					.getCustomData(turkov.util.TUClusterUtil.SUPPLEMENTAL_ITEM) != null)
				continue;
			//System.out.println("processing sequence " + i + "...");
			// here, k are the global id of elements
			Vector k = new Vector();
			final Object[] currentSequence = (Object[]) ((TUClusterableItem) cluster
					.get(i)).getValue();

			final Object[] longestCommonSubsequence = TULCS
					.getLongestCommonSubsequence(outlierSequence,
							currentSequence);

			//System.out.println("The LCS is: " + TUStringBuilder.concatenate(longestCommonSubsequence));

			// to determine mappings below
			int currentPosition = 0;

			int[] lcsToCurrentSequenceIndices = new int[longestCommonSubsequence.length];

			int[] lcsToOutlierSequenceIndices = new int[longestCommonSubsequence.length];

			int[] outlierSequenceToLcsIndices = new int[outlierSequence.length];

			// first determine lcsToCurrentSequenceMapping
			for (int j = 0; j < longestCommonSubsequence.length; j++)
			{
				lcsToCurrentSequenceIndices[j] = TUCollectionUtil.getPosition(
						longestCommonSubsequence[j], currentSequence,
						currentPosition);

				currentPosition = lcsToCurrentSequenceIndices[j] + 1;
			}

			//TUStringBuilder.printCollection(longestCommonSubsequence);

			//reset currentPosition
			currentPosition = 0;

			// second, determine outlierSequenceToLcsMapping
			for (int j = 0; j < outlierSequence.length; j++)
			{
				outlierSequenceToLcsIndices[j] = TUCollectionUtil.getPosition(
						outlierSequence[j], longestCommonSubsequence,
						currentPosition);

				currentPosition = outlierSequenceToLcsIndices[j] + 1;
			}

			//third... lcstooutlier
			//StringBuffer sb = new StringBuffer();
			currentPosition = 0;
			for (int j = 0; j < longestCommonSubsequence.length; j++)
			{
				lcsToOutlierSequenceIndices[j] = TUCollectionUtil.getPosition(
						longestCommonSubsequence[j], outlierSequence,
						currentPosition);
				//System.err.println(j + " --> " + lcsToOutlierSequenceIndices[j]);
				//System.err.println("the outlier sequence is: " + TUStringBuilder.concatenate(outlierSequence));
				//sb.append(j + " --> " + lcsToOutlierSequenceIndices[j]);
				//sb.append("the outlier sequence is: " + TUStringBuilder.concatenate(outlierSequence));
				//sb.append(System.getProperty("line.separator"));
				currentPosition = lcsToOutlierSequenceIndices[j] + 1;
			}

			/*
			 no lcs means add everything
			 */
			if (longestCommonSubsequence.length == 0)
			{
				for (int currentSequenceIndex = 0; currentSequenceIndex < currentSequence.length; currentSequenceIndex++)
				{
					int elementId = 0;

					//try
					//{
					elementId = ((Integer) (uniqueElementsToIds
							.get(currentSequence[currentSequenceIndex])))
							.intValue();
					//}

					/*
					catch(NullPointerException e)
					{
					uniqueElementsToIds.put(currentSequence[currentSequenceIndex], )
					}
					 */

					b[0][elementId] += ((double) longestCommonSubsequence.length + 1)
							/ ((double) currentSequence.length + 1);

					//System.out.println(b[0][elementId]);
				}
			}

			for (int lcsIndex = 0; lcsIndex < longestCommonSubsequence.length; lcsIndex++)
			{
				/*
				 first check: the first item in the LCS is not the first item in
				 the current sequence. need to add everything from point
				 zero in the current sequence up to (but not including the
				 first item in the LCS
				 */
				if (lcsIndex == 0 && lcsToCurrentSequenceIndices[lcsIndex] != 0)
				{
					for (int currentSequenceIndex = 0; currentSequenceIndex < lcsToCurrentSequenceIndices[lcsIndex]; currentSequenceIndex++)
					{
						int elementId = ((Integer) (uniqueElementsToIds
								.get(currentSequence[currentSequenceIndex])))
								.intValue();

						//System.out.println("lcs index is: " + lcsIndex);
						b[lcsToOutlierSequenceIndices[lcsIndex]][elementId] += ((double) longestCommonSubsequence.length + 1)
								/ ((double) currentSequence.length + 1);
					}
				}

				/*
				 second check: the last item in the LCS is not the last item in
				 the current sequence. need to add everything from the current
				 sequence 'till the end of the current sequence starting from where
				 there is no overlap between the LCS and the current sequence.
				 */
				if (lcsIndex == longestCommonSubsequence.length - 1
						&& lcsToCurrentSequenceIndices[lcsIndex] != currentSequence.length - 1)
				{
					for (int currentSequenceIndex = lcsToCurrentSequenceIndices[lcsIndex] + 1; currentSequenceIndex < currentSequence.length; currentSequenceIndex++)
					{
						int elementId = ((Integer) (uniqueElementsToIds
								.get(currentSequence[currentSequenceIndex])))
								.intValue();

						//System.out.println("lcsToOutlierSequenceIndices[lcsIndex] = " + lcsToOutlierSequenceIndices[lcsIndex]);
						//System.out.println("and item is: " + currentSequence[currentSequenceIndex]);
						b[lcsToOutlierSequenceIndices[lcsIndex] + 1][elementId] += ((double) longestCommonSubsequence.length + 1)
								/ ((double) currentSequence.length + 1);
					}
				}

				/*
				 third check: if the current item is not the first item in the LCS
				 then see if there is anything missing between this current LCS item
				 and the previous LCS item. Add all elements which are missing
				 (according to the current sequence) where overlap between the two
				 surrounding LCS elements does not occurr.
				 */
				if (lcsIndex != 0
						&& (lcsToCurrentSequenceIndices[lcsIndex] - lcsToCurrentSequenceIndices[lcsIndex - 1]) > 1)
				{
					int endLocation = lcsToCurrentSequenceIndices[lcsIndex] - 1;
					int startLocation = lcsToCurrentSequenceIndices[lcsIndex - 1] + 1;

					for (int currentSequenceIndex = startLocation; currentSequenceIndex <= endLocation; currentSequenceIndex++)
					{
						int elementId = ((Integer) (uniqueElementsToIds
								.get(currentSequence[currentSequenceIndex])))
								.intValue();

						//try
						{
							b[lcsToOutlierSequenceIndices[lcsIndex]][elementId] += ((double) longestCommonSubsequence.length + 1)
									/ ((double) currentSequence.length + 1);
						}
						/*
						catch(ArrayIndexOutOfBoundsException e)
						{
						    
						    System.err.println("the element id is: " + elementId);
						    System.err.println("the lcs index is: " + lcsIndex);  
						    System.err.println("b is: " + b.length + " X " + b[0].length);
						    System.err.println("lcsToOutlierSequenceIndices[lcsIndex]: " + lcsToOutlierSequenceIndices[lcsIndex]);
						    System.err.println("the lcs is: " + TUStringBuilder.concatenate(longestCommonSubsequence, " "));
						    System.err.println("the current sequence is: " + TUStringBuilder.concatenate(currentSequence," "));
						    //System.err.println(sb);
						    System.exit(0);
						}
						 */
					}
				}
			}

			int lengthOfLongestCommonSubsequence = -1;

			final TUClusterableItem currentSequenceObject = ((TUClusterableItem) cluster
					.get(i));

			if (currentSequenceObject.getCustomData(DISTANCE_FROM_MEDOID) == null)
			{
				lengthOfLongestCommonSubsequence = TULCS
						.getLengthOfLongestCommonSubsequence(consensusSequence,
								currentSequence);

				currentSequenceObject.setCustomData(DISTANCE_FROM_MEDOID,
						new Integer(lengthOfLongestCommonSubsequence));
			}

			else
				lengthOfLongestCommonSubsequence = ((Integer) currentSequenceObject
						.getCustomData(DISTANCE_FROM_MEDOID)).intValue();

			score += (longestCommonSubsequence.length * lengthOfLongestCommonSubsequence)
					/ (Math.sqrt(outlierSequence.length) * currentSequence.length);
		}

		Vector maxPoints = new Vector();
		Vector resultScore = new Vector();
		while (true)
		{
			final double tempMax = TUMaxMin.getMaxValue(b);
			maxPoints.clear();
			for (int i = 0; i < b.length; i++)
			{
				//System.out.println("now temp max is: " + tempMax);
				for (int j = 0; j < b[i].length; j++)
				{
					if (b[i][j] == tempMax)
					{
						maxPoints.add(new Point(i, j));
						/*
						 System.out.println("adding item : " + idsToUniqueElements.get(new Integer(j)));
						 System.out.println("at location: " + i);
						 */
					}
					/*
					System.out.println("temp max for " + idsToUniqueElements.get(new Integer(j)) 
					+ " is : " + b[i][j]);
					 */
				}
			}

			final int k = maxPoints.size();

			oldScore = score;

			score = (1 / Math.sqrt(outlierSequence.length + k))
					* (Math.sqrt(outlierSequence.length) * score + k * tempMax);

			/*
			 score = (1 / Math.sqrt(outlierSequence.length + k))
			 * ((1 / Math.sqrt(outlierSequence.length))
			 * score
			 + k
			 * tempMax);            
			 */
			/*
			System.out.println("old score: " + oldScore);
			System.out.println("new score: " + score);            
			 */

			if (score > oldScore
					|| (Double.isNaN(score) && Double.isNaN(oldScore)))
			{
				for (int i = 0; i < maxPoints.size(); i++)
				{
					int x = ((Point) maxPoints.get(i)).x;
					int y = ((Point) maxPoints.get(i)).y;

					missingCharacterIndeciesAndIds.add(maxPoints.get(i));
					resultScore.add(new Double(b[x][y]));
					b[x][y] = 0;
				}

			}

			if (score <= oldScore
					|| (Double.isNaN(score) && Double.isNaN(oldScore)))
			{
				break;
			}
			oldScore = score;
		}

		TUSequenceAlteration[] result = new TUSequenceAlteration[missingCharacterIndeciesAndIds
				.size()];

		for (int i = 0; i < missingCharacterIndeciesAndIds.size(); i++)
		{
			Point p = (Point) missingCharacterIndeciesAndIds.get(i);
			result[i] = new TUSequenceAlteration(p.x, idsToUniqueElements
					.get(new Integer(p.y)), ((Double) resultScore.get(i))
					.doubleValue());
		}

		return result;
	}

	public static TUSequenceAlteration[] getMissingElementsInOrder(
			final Object[] incomingOutlierSequence, final TUCluster cluster,
			final HashMap uniqueElementsToIds,
			final HashMap idsToUniqueElements, int sequenceId)
	{
		/*
		TUObjectExtension[] outlierSequence = new TUObjectExtension[incomingOutlierSequence.length];
		for(int i = 0; i < outlierSequence.length; i++)
		{
		    TUObjectExtension objectExtension = new TUObjectExtension(incomingOutlierSequence[i]);
		    objectExtension.setProperty("originalLocation", new Integer(i));
		    outlierSequence[i] = objectExtension;
		}
		 */
		/*
		System.err.println("=====================");
		System.err.println("the outlier sequence is: " + TUStringBuilder.concatenate(outlierSequence));

		Object medoid = cluster.getMedoid(TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE);
		Object[] stuff = (Object[])((TUClusterableItem)medoid).getValue();
		
		
		System.err.println("the medoid is: "
		        + TUStringBuilder.concatenate(stuff));
		 */
		// get what's missing 
		TUSequenceAlteration[] suggestedAdditions = getMissingElements(
				incomingOutlierSequence, cluster, uniqueElementsToIds,
				idsToUniqueElements);

		for (int i = 0; i < suggestedAdditions.length; i++)
		{
			//System.err.println(suggestedAdditions[i].toString() + " before position " + suggestedAdditions[i].getIndex());
			Object object = suggestedAdditions[i].getObject();
			TUObjectExtension objectExtension = new TUObjectExtension(object);

			objectExtension.setProperty("insertionScore", new Double(
					suggestedAdditions[i].getScore()));

			objectExtension.setProperty("insertionPosition", new Integer(
					suggestedAdditions[i].getIndex()));

			suggestedAdditions[i].setObject(objectExtension);
		}

		// apply the suggested additions
		Object[] tempOutlierSequence = TUSequenceUtil.applyAdditions(
				incomingOutlierSequence, suggestedAdditions);

		//System.err.println("after add modification the result is: " + TUStringBuilder.concatenate(tempOutlierSequence));

		// after applying the suggested insertions, see if anything needs to be removed
		TUSequenceAlteration[] neccessaryRemovals = getNonEssentialElements(
				tempOutlierSequence, cluster, uniqueElementsToIds,
				idsToUniqueElements, sequenceId);

		//System.err.println(TUStringBuilder.concatenate(neccessaryRemovals, " neccessaryRemoval "));

		// apply the removals
		//System.err.println("THIS MANY REMOVALS TO BE MADE: " + neccessaryRemovals.length);

		//System.err.println("before REMOVALS the result is: " + TUStringBuilder.concatenate(tempOutlierSequence));

		Object[] convertedOutlierSequence = TUSequenceUtil.applyRemovals(
				tempOutlierSequence, neccessaryRemovals);

		//System.err.println("after REMOVALS the result is: " + TUStringBuilder.concatenate(convertedOutlierSequence));

		// now convertedOutlierSequence should look like the medoid of the cluster
		/*
		System.err.println("calculated result sequence is: "
		        + TUStringBuilder.concatenate(convertedOutlierSequence));
		 */
		// now go through the result sequence and see if there are any "extended"
		// objects in there. This is the sequenceAlteration result to return.
		Vector finalAdditions = new Vector();

		for (int i = 0; i < convertedOutlierSequence.length; i++)
		{
			if (convertedOutlierSequence[i] instanceof TUObjectExtension)
			{
				TUObjectExtension item = (TUObjectExtension) convertedOutlierSequence[i];
				// found an item that was originailly inersted at one point
				finalAdditions.add(new TUSequenceAlteration(((Integer) item
						.getPropertyValue("insertionPosition")).intValue(),
						item.getRootObject(), ((Double) item
								.getPropertyValue("insertionScore"))
								.doubleValue()));
			}
		}

		TUSequenceAlteration[] result = new TUSequenceAlteration[finalAdditions
				.size()];
		for (int i = 0; i < result.length; i++)
		{
			result[i] = (TUSequenceAlteration) finalAdditions.get(i);
		}

		//System.err.println("=====================");
		return result;

	}

	public final static String SUPPLEMENTAL_ITEM = "SUPPLEMENTAL_ITEM";
	public final static String TRUE = "TRUE";
	public final static String OUTLIER = "OUTLIER";
	public final static String DISTANCE_FROM_MEDOID = "DISTANCE_FROM_MEDOID";
}
